# Generated from ./PlantUML.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,44,259,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,2,18,7,18,2,19,7,19,2,20,
        7,20,2,21,7,21,2,22,7,22,2,23,7,23,2,24,7,24,1,0,1,0,1,0,5,0,54,
        8,0,10,0,12,0,57,9,0,1,0,1,0,1,1,1,1,1,1,1,1,3,1,65,8,1,1,1,4,1,
        68,8,1,11,1,12,1,69,1,2,1,2,1,2,1,2,1,2,1,3,1,3,3,3,79,8,3,1,3,1,
        3,3,3,83,8,3,1,3,1,3,5,3,87,8,3,10,3,12,3,90,9,3,1,3,1,3,5,3,94,
        8,3,10,3,12,3,97,9,3,1,3,1,3,1,4,1,4,3,4,103,8,4,1,5,1,5,3,5,107,
        8,5,1,6,1,6,3,6,111,8,6,1,6,1,6,1,6,1,6,3,6,117,8,6,1,6,3,6,120,
        8,6,1,6,1,6,1,6,3,6,125,8,6,1,7,1,7,1,8,3,8,130,8,8,1,8,1,8,3,8,
        134,8,8,1,9,3,9,137,8,9,1,9,3,9,140,8,9,3,9,142,8,9,1,9,1,9,3,9,
        146,8,9,1,9,3,9,149,8,9,3,9,151,8,9,1,10,3,10,154,8,10,1,10,3,10,
        157,8,10,3,10,159,8,10,1,10,1,10,3,10,163,8,10,1,10,3,10,166,8,10,
        3,10,168,8,10,1,11,1,11,1,11,3,11,173,8,11,1,11,1,11,1,12,1,12,1,
        12,1,13,1,13,1,13,1,13,3,13,184,8,13,1,13,1,13,1,14,1,14,1,15,3,
        15,191,8,15,1,15,1,15,1,15,1,15,1,15,1,16,3,16,199,8,16,1,16,3,16,
        202,8,16,1,16,1,16,1,16,1,16,1,16,3,16,209,8,16,3,16,211,8,16,1,
        16,1,16,1,16,3,16,216,8,16,1,16,1,16,1,17,1,17,1,17,1,17,1,17,3,
        17,225,8,17,1,18,3,18,228,8,18,1,18,1,18,3,18,232,8,18,1,19,1,19,
        3,19,236,8,19,1,20,1,20,1,20,1,20,1,20,5,20,243,8,20,10,20,12,20,
        246,9,20,1,20,1,20,1,21,1,21,1,21,1,22,1,22,1,23,1,23,1,24,1,24,
        1,24,0,0,25,0,2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,
        38,40,42,44,46,48,0,5,2,0,39,39,41,41,1,0,38,40,1,0,21,24,1,0,25,
        33,1,0,34,35,278,0,50,1,0,0,0,2,64,1,0,0,0,4,71,1,0,0,0,6,78,1,0,
        0,0,8,100,1,0,0,0,10,106,1,0,0,0,12,108,1,0,0,0,14,126,1,0,0,0,16,
        129,1,0,0,0,18,141,1,0,0,0,20,158,1,0,0,0,22,169,1,0,0,0,24,176,
        1,0,0,0,26,179,1,0,0,0,28,187,1,0,0,0,30,190,1,0,0,0,32,198,1,0,
        0,0,34,219,1,0,0,0,36,227,1,0,0,0,38,235,1,0,0,0,40,237,1,0,0,0,
        42,249,1,0,0,0,44,252,1,0,0,0,46,254,1,0,0,0,48,256,1,0,0,0,50,51,
        5,36,0,0,51,55,5,43,0,0,52,54,3,2,1,0,53,52,1,0,0,0,54,57,1,0,0,
        0,55,53,1,0,0,0,55,56,1,0,0,0,56,58,1,0,0,0,57,55,1,0,0,0,58,59,
        5,37,0,0,59,1,1,0,0,0,60,65,3,4,2,0,61,65,3,6,3,0,62,65,3,10,5,0,
        63,65,3,40,20,0,64,60,1,0,0,0,64,61,1,0,0,0,64,62,1,0,0,0,64,63,
        1,0,0,0,65,67,1,0,0,0,66,68,5,43,0,0,67,66,1,0,0,0,68,69,1,0,0,0,
        69,67,1,0,0,0,69,70,1,0,0,0,70,3,1,0,0,0,71,72,5,1,0,0,72,73,5,2,
        0,0,73,74,5,39,0,0,74,75,5,43,0,0,75,5,1,0,0,0,76,79,3,8,4,0,77,
        79,5,3,0,0,78,76,1,0,0,0,78,77,1,0,0,0,79,80,1,0,0,0,80,82,5,38,
        0,0,81,83,3,24,12,0,82,81,1,0,0,0,82,83,1,0,0,0,83,84,1,0,0,0,84,
        88,5,4,0,0,85,87,5,43,0,0,86,85,1,0,0,0,87,90,1,0,0,0,88,86,1,0,
        0,0,88,89,1,0,0,0,89,95,1,0,0,0,90,88,1,0,0,0,91,94,3,30,15,0,92,
        94,3,32,16,0,93,91,1,0,0,0,93,92,1,0,0,0,94,97,1,0,0,0,95,93,1,0,
        0,0,95,96,1,0,0,0,96,98,1,0,0,0,97,95,1,0,0,0,98,99,5,5,0,0,99,7,
        1,0,0,0,100,102,5,6,0,0,101,103,5,3,0,0,102,101,1,0,0,0,102,103,
        1,0,0,0,103,9,1,0,0,0,104,107,3,12,6,0,105,107,3,22,11,0,106,104,
        1,0,0,0,106,105,1,0,0,0,107,11,1,0,0,0,108,110,5,38,0,0,109,111,
        3,26,13,0,110,109,1,0,0,0,110,111,1,0,0,0,111,116,1,0,0,0,112,117,
        3,14,7,0,113,117,3,16,8,0,114,117,3,18,9,0,115,117,3,20,10,0,116,
        112,1,0,0,0,116,113,1,0,0,0,116,114,1,0,0,0,116,115,1,0,0,0,117,
        119,1,0,0,0,118,120,3,26,13,0,119,118,1,0,0,0,119,120,1,0,0,0,120,
        121,1,0,0,0,121,124,5,38,0,0,122,123,5,7,0,0,123,125,5,38,0,0,124,
        122,1,0,0,0,124,125,1,0,0,0,125,13,1,0,0,0,126,127,5,8,0,0,127,15,
        1,0,0,0,128,130,5,9,0,0,129,128,1,0,0,0,129,130,1,0,0,0,130,131,
        1,0,0,0,131,133,5,8,0,0,132,134,5,10,0,0,133,132,1,0,0,0,133,134,
        1,0,0,0,134,17,1,0,0,0,135,137,5,11,0,0,136,135,1,0,0,0,136,137,
        1,0,0,0,137,142,1,0,0,0,138,140,5,9,0,0,139,138,1,0,0,0,139,140,
        1,0,0,0,140,142,1,0,0,0,141,136,1,0,0,0,141,139,1,0,0,0,142,143,
        1,0,0,0,143,150,5,8,0,0,144,146,5,10,0,0,145,144,1,0,0,0,145,146,
        1,0,0,0,146,151,1,0,0,0,147,149,5,11,0,0,148,147,1,0,0,0,148,149,
        1,0,0,0,149,151,1,0,0,0,150,145,1,0,0,0,150,148,1,0,0,0,151,19,1,
        0,0,0,152,154,5,41,0,0,153,152,1,0,0,0,153,154,1,0,0,0,154,159,1,
        0,0,0,155,157,5,9,0,0,156,155,1,0,0,0,156,157,1,0,0,0,157,159,1,
        0,0,0,158,153,1,0,0,0,158,156,1,0,0,0,159,160,1,0,0,0,160,167,5,
        8,0,0,161,163,5,10,0,0,162,161,1,0,0,0,162,163,1,0,0,0,163,168,1,
        0,0,0,164,166,5,41,0,0,165,164,1,0,0,0,165,166,1,0,0,0,166,168,1,
        0,0,0,167,162,1,0,0,0,167,165,1,0,0,0,168,21,1,0,0,0,169,172,5,38,
        0,0,170,173,5,12,0,0,171,173,5,13,0,0,172,170,1,0,0,0,172,171,1,
        0,0,0,173,174,1,0,0,0,174,175,5,38,0,0,175,23,1,0,0,0,176,177,5,
        14,0,0,177,178,5,38,0,0,178,25,1,0,0,0,179,180,5,44,0,0,180,183,
        3,28,14,0,181,182,5,15,0,0,182,184,3,28,14,0,183,181,1,0,0,0,183,
        184,1,0,0,0,184,185,1,0,0,0,185,186,5,44,0,0,186,27,1,0,0,0,187,
        188,7,0,0,0,188,29,1,0,0,0,189,191,3,44,22,0,190,189,1,0,0,0,190,
        191,1,0,0,0,191,192,1,0,0,0,192,193,5,38,0,0,193,194,5,7,0,0,194,
        195,3,38,19,0,195,196,5,43,0,0,196,31,1,0,0,0,197,199,3,44,22,0,
        198,197,1,0,0,0,198,199,1,0,0,0,199,201,1,0,0,0,200,202,3,48,24,
        0,201,200,1,0,0,0,201,202,1,0,0,0,202,203,1,0,0,0,203,204,5,38,0,
        0,204,210,5,16,0,0,205,208,3,34,17,0,206,207,5,17,0,0,207,209,3,
        34,17,0,208,206,1,0,0,0,208,209,1,0,0,0,209,211,1,0,0,0,210,205,
        1,0,0,0,210,211,1,0,0,0,211,212,1,0,0,0,212,215,5,18,0,0,213,214,
        5,7,0,0,214,216,3,38,19,0,215,213,1,0,0,0,215,216,1,0,0,0,216,217,
        1,0,0,0,217,218,5,43,0,0,218,33,1,0,0,0,219,220,5,38,0,0,220,221,
        5,7,0,0,221,224,3,38,19,0,222,223,5,19,0,0,223,225,3,36,18,0,224,
        222,1,0,0,0,224,225,1,0,0,0,225,35,1,0,0,0,226,228,5,44,0,0,227,
        226,1,0,0,0,227,228,1,0,0,0,228,229,1,0,0,0,229,231,7,1,0,0,230,
        232,5,44,0,0,231,230,1,0,0,0,231,232,1,0,0,0,232,37,1,0,0,0,233,
        236,3,46,23,0,234,236,5,38,0,0,235,233,1,0,0,0,235,234,1,0,0,0,236,
        39,1,0,0,0,237,238,5,20,0,0,238,239,5,38,0,0,239,240,5,4,0,0,240,
        244,5,43,0,0,241,243,3,42,21,0,242,241,1,0,0,0,243,246,1,0,0,0,244,
        242,1,0,0,0,244,245,1,0,0,0,245,247,1,0,0,0,246,244,1,0,0,0,247,
        248,5,5,0,0,248,41,1,0,0,0,249,250,5,38,0,0,250,251,5,43,0,0,251,
        43,1,0,0,0,252,253,7,2,0,0,253,45,1,0,0,0,254,255,7,3,0,0,255,47,
        1,0,0,0,256,257,7,4,0,0,257,49,1,0,0,0,41,55,64,69,78,82,88,93,95,
        102,106,110,116,119,124,129,133,136,139,141,145,148,150,153,156,
        158,162,165,167,172,183,190,198,201,208,210,215,224,227,231,235,
        244
    ]

class PlantUMLParser ( Parser ):

    grammarFileName = "PlantUML.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'skinparam'", "'groupInheritance'", "'class'", 
                     "'{'", "'}'", "'abstract'", "':'", "'--'", "'<'", "'>'", 
                     "'o'", "'<|--'", "'--|>'", "'extends'", "'..'", "'('", 
                     "','", "')'", "'='", "'enum'", "'#'", "'-'", "'~'", 
                     "'+'", "'int'", "'float'", "'str'", "'string'", "'bool'", 
                     "'time'", "'date'", "'datetime'", "'timedelta'", "'{static}'", 
                     "'{abstract}'", "'@startuml'", "'@enduml'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'*'", "<INVALID>", "<INVALID>", 
                     "'\"'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "Start", "End", "ID", "INT", "FLOAT", "ASTK", "WS", 
                      "NL", "D_QUOTE" ]

    RULE_domainModel = 0
    RULE_element = 1
    RULE_skinParam = 2
    RULE_class = 3
    RULE_abstract = 4
    RULE_relationship = 5
    RULE_association = 6
    RULE_bidirectional = 7
    RULE_unidirectional = 8
    RULE_aggregation = 9
    RULE_composition = 10
    RULE_inheritance = 11
    RULE_extends = 12
    RULE_cardinality = 13
    RULE_cardinalityVal = 14
    RULE_attribute = 15
    RULE_method = 16
    RULE_parameter = 17
    RULE_value = 18
    RULE_dType = 19
    RULE_enumeration = 20
    RULE_enumLiteral = 21
    RULE_visibility = 22
    RULE_primitiveData = 23
    RULE_modifier = 24

    ruleNames =  [ "domainModel", "element", "skinParam", "class", "abstract", 
                   "relationship", "association", "bidirectional", "unidirectional", 
                   "aggregation", "composition", "inheritance", "extends", 
                   "cardinality", "cardinalityVal", "attribute", "method", 
                   "parameter", "value", "dType", "enumeration", "enumLiteral", 
                   "visibility", "primitiveData", "modifier" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    T__27=28
    T__28=29
    T__29=30
    T__30=31
    T__31=32
    T__32=33
    T__33=34
    T__34=35
    Start=36
    End=37
    ID=38
    INT=39
    FLOAT=40
    ASTK=41
    WS=42
    NL=43
    D_QUOTE=44

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class DomainModelContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Start(self):
            return self.getToken(PlantUMLParser.Start, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def End(self):
            return self.getToken(PlantUMLParser.End, 0)

        def element(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.ElementContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.ElementContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_domainModel

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDomainModel" ):
                listener.enterDomainModel(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDomainModel" ):
                listener.exitDomainModel(self)




    def domainModel(self):

        localctx = PlantUMLParser.DomainModelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_domainModel)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.match(PlantUMLParser.Start)
            self.state = 51
            self.match(PlantUMLParser.NL)
            self.state = 55
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 274878955594) != 0):
                self.state = 52
                self.element()
                self.state = 57
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 58
            self.match(PlantUMLParser.End)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ElementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def skinParam(self):
            return self.getTypedRuleContext(PlantUMLParser.SkinParamContext,0)


        def class_(self):
            return self.getTypedRuleContext(PlantUMLParser.ClassContext,0)


        def relationship(self):
            return self.getTypedRuleContext(PlantUMLParser.RelationshipContext,0)


        def enumeration(self):
            return self.getTypedRuleContext(PlantUMLParser.EnumerationContext,0)


        def NL(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.NL)
            else:
                return self.getToken(PlantUMLParser.NL, i)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_element

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterElement" ):
                listener.enterElement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitElement" ):
                listener.exitElement(self)




    def element(self):

        localctx = PlantUMLParser.ElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_element)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 64
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1]:
                self.state = 60
                self.skinParam()
                pass
            elif token in [3, 6]:
                self.state = 61
                self.class_()
                pass
            elif token in [38]:
                self.state = 62
                self.relationship()
                pass
            elif token in [20]:
                self.state = 63
                self.enumeration()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 67 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 66
                self.match(PlantUMLParser.NL)
                self.state = 69 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==43):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SkinParamContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(PlantUMLParser.INT, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_skinParam

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSkinParam" ):
                listener.enterSkinParam(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSkinParam" ):
                listener.exitSkinParam(self)




    def skinParam(self):

        localctx = PlantUMLParser.SkinParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_skinParam)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 71
            self.match(PlantUMLParser.T__0)
            self.state = 72
            self.match(PlantUMLParser.T__1)
            self.state = 73
            self.match(PlantUMLParser.INT)
            self.state = 74
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ClassContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def abstract(self):
            return self.getTypedRuleContext(PlantUMLParser.AbstractContext,0)


        def extends(self):
            return self.getTypedRuleContext(PlantUMLParser.ExtendsContext,0)


        def NL(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.NL)
            else:
                return self.getToken(PlantUMLParser.NL, i)

        def attribute(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.AttributeContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.AttributeContext,i)


        def method(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.MethodContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.MethodContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_class

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClass" ):
                listener.enterClass(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClass" ):
                listener.exitClass(self)




    def class_(self):

        localctx = PlantUMLParser.ClassContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_class)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [6]:
                self.state = 76
                self.abstract()
                pass
            elif token in [3]:
                self.state = 77
                self.match(PlantUMLParser.T__2)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 80
            self.match(PlantUMLParser.ID)
            self.state = 82
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==14:
                self.state = 81
                self.extends()


            self.state = 84
            self.match(PlantUMLParser.T__3)
            self.state = 88
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==43:
                self.state = 85
                self.match(PlantUMLParser.NL)
                self.state = 90
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 95
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 326448971776) != 0):
                self.state = 93
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,6,self._ctx)
                if la_ == 1:
                    self.state = 91
                    self.attribute()
                    pass

                elif la_ == 2:
                    self.state = 92
                    self.method()
                    pass


                self.state = 97
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 98
            self.match(PlantUMLParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AbstractContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_abstract

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAbstract" ):
                listener.enterAbstract(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAbstract" ):
                listener.exitAbstract(self)




    def abstract(self):

        localctx = PlantUMLParser.AbstractContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_abstract)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 100
            self.match(PlantUMLParser.T__5)
            self.state = 102
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==3:
                self.state = 101
                self.match(PlantUMLParser.T__2)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RelationshipContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def association(self):
            return self.getTypedRuleContext(PlantUMLParser.AssociationContext,0)


        def inheritance(self):
            return self.getTypedRuleContext(PlantUMLParser.InheritanceContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_relationship

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRelationship" ):
                listener.enterRelationship(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRelationship" ):
                listener.exitRelationship(self)




    def relationship(self):

        localctx = PlantUMLParser.RelationshipContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_relationship)
        try:
            self.state = 106
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 104
                self.association()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 105
                self.inheritance()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AssociationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.c_left = None # CardinalityContext
            self.c_right = None # CardinalityContext

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ID)
            else:
                return self.getToken(PlantUMLParser.ID, i)

        def bidirectional(self):
            return self.getTypedRuleContext(PlantUMLParser.BidirectionalContext,0)


        def unidirectional(self):
            return self.getTypedRuleContext(PlantUMLParser.UnidirectionalContext,0)


        def aggregation(self):
            return self.getTypedRuleContext(PlantUMLParser.AggregationContext,0)


        def composition(self):
            return self.getTypedRuleContext(PlantUMLParser.CompositionContext,0)


        def cardinality(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.CardinalityContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.CardinalityContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_association

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAssociation" ):
                listener.enterAssociation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAssociation" ):
                listener.exitAssociation(self)




    def association(self):

        localctx = PlantUMLParser.AssociationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_association)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 108
            self.match(PlantUMLParser.ID)
            self.state = 110
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==44:
                self.state = 109
                localctx.c_left = self.cardinality()


            self.state = 116
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
            if la_ == 1:
                self.state = 112
                self.bidirectional()
                pass

            elif la_ == 2:
                self.state = 113
                self.unidirectional()
                pass

            elif la_ == 3:
                self.state = 114
                self.aggregation()
                pass

            elif la_ == 4:
                self.state = 115
                self.composition()
                pass


            self.state = 119
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==44:
                self.state = 118
                localctx.c_right = self.cardinality()


            self.state = 121
            self.match(PlantUMLParser.ID)
            self.state = 124
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==7:
                self.state = 122
                self.match(PlantUMLParser.T__6)
                self.state = 123
                self.match(PlantUMLParser.ID)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BidirectionalContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_bidirectional

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBidirectional" ):
                listener.enterBidirectional(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBidirectional" ):
                listener.exitBidirectional(self)




    def bidirectional(self):

        localctx = PlantUMLParser.BidirectionalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_bidirectional)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 126
            self.match(PlantUMLParser.T__7)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UnidirectionalContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.nav_l = None # Token
            self.nav_r = None # Token


        def getRuleIndex(self):
            return PlantUMLParser.RULE_unidirectional

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnidirectional" ):
                listener.enterUnidirectional(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnidirectional" ):
                listener.exitUnidirectional(self)




    def unidirectional(self):

        localctx = PlantUMLParser.UnidirectionalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_unidirectional)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 129
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==9:
                self.state = 128
                localctx.nav_l = self.match(PlantUMLParser.T__8)


            self.state = 131
            self.match(PlantUMLParser.T__7)
            self.state = 133
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==10:
                self.state = 132
                localctx.nav_r = self.match(PlantUMLParser.T__9)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AggregationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.aggr_l = None # Token
            self.aggr_r = None # Token


        def getRuleIndex(self):
            return PlantUMLParser.RULE_aggregation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAggregation" ):
                listener.enterAggregation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAggregation" ):
                listener.exitAggregation(self)




    def aggregation(self):

        localctx = PlantUMLParser.AggregationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_aggregation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 141
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
            if la_ == 1:
                self.state = 136
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 135
                    localctx.aggr_l = self.match(PlantUMLParser.T__10)


                pass

            elif la_ == 2:
                self.state = 139
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==9:
                    self.state = 138
                    self.match(PlantUMLParser.T__8)


                pass


            self.state = 143
            self.match(PlantUMLParser.T__7)
            self.state = 150
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,21,self._ctx)
            if la_ == 1:
                self.state = 145
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==10:
                    self.state = 144
                    self.match(PlantUMLParser.T__9)


                pass

            elif la_ == 2:
                self.state = 148
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 147
                    localctx.aggr_r = self.match(PlantUMLParser.T__10)


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CompositionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.comp_l = None # Token
            self.comp_r = None # Token

        def ASTK(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ASTK)
            else:
                return self.getToken(PlantUMLParser.ASTK, i)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_composition

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComposition" ):
                listener.enterComposition(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComposition" ):
                listener.exitComposition(self)




    def composition(self):

        localctx = PlantUMLParser.CompositionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_composition)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 158
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,24,self._ctx)
            if la_ == 1:
                self.state = 153
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==41:
                    self.state = 152
                    localctx.comp_l = self.match(PlantUMLParser.ASTK)


                pass

            elif la_ == 2:
                self.state = 156
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==9:
                    self.state = 155
                    self.match(PlantUMLParser.T__8)


                pass


            self.state = 160
            self.match(PlantUMLParser.T__7)
            self.state = 167
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,27,self._ctx)
            if la_ == 1:
                self.state = 162
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==10:
                    self.state = 161
                    self.match(PlantUMLParser.T__9)


                pass

            elif la_ == 2:
                self.state = 165
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==41:
                    self.state = 164
                    localctx.comp_r = self.match(PlantUMLParser.ASTK)


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class InheritanceContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.inh_left = None # Token

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.ID)
            else:
                return self.getToken(PlantUMLParser.ID, i)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_inheritance

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInheritance" ):
                listener.enterInheritance(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInheritance" ):
                listener.exitInheritance(self)




    def inheritance(self):

        localctx = PlantUMLParser.InheritanceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_inheritance)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 169
            self.match(PlantUMLParser.ID)
            self.state = 172
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [12]:
                self.state = 170
                localctx.inh_left = self.match(PlantUMLParser.T__11)
                pass
            elif token in [13]:
                self.state = 171
                self.match(PlantUMLParser.T__12)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 174
            self.match(PlantUMLParser.ID)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ExtendsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_extends

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExtends" ):
                listener.enterExtends(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExtends" ):
                listener.exitExtends(self)




    def extends(self):

        localctx = PlantUMLParser.ExtendsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_extends)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 176
            self.match(PlantUMLParser.T__13)
            self.state = 177
            self.match(PlantUMLParser.ID)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CardinalityContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.min_ = None # CardinalityValContext
            self.max_ = None # CardinalityValContext

        def D_QUOTE(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.D_QUOTE)
            else:
                return self.getToken(PlantUMLParser.D_QUOTE, i)

        def cardinalityVal(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.CardinalityValContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.CardinalityValContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_cardinality

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCardinality" ):
                listener.enterCardinality(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCardinality" ):
                listener.exitCardinality(self)




    def cardinality(self):

        localctx = PlantUMLParser.CardinalityContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_cardinality)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 179
            self.match(PlantUMLParser.D_QUOTE)
            self.state = 180
            localctx.min_ = self.cardinalityVal()
            self.state = 183
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 181
                self.match(PlantUMLParser.T__14)
                self.state = 182
                localctx.max_ = self.cardinalityVal()


            self.state = 185
            self.match(PlantUMLParser.D_QUOTE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CardinalityValContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(PlantUMLParser.INT, 0)

        def ASTK(self):
            return self.getToken(PlantUMLParser.ASTK, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_cardinalityVal

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCardinalityVal" ):
                listener.enterCardinalityVal(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCardinalityVal" ):
                listener.exitCardinalityVal(self)




    def cardinalityVal(self):

        localctx = PlantUMLParser.CardinalityValContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_cardinalityVal)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 187
            _la = self._input.LA(1)
            if not(_la==39 or _la==41):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AttributeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def dType(self):
            return self.getTypedRuleContext(PlantUMLParser.DTypeContext,0)


        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def visibility(self):
            return self.getTypedRuleContext(PlantUMLParser.VisibilityContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_attribute

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAttribute" ):
                listener.enterAttribute(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAttribute" ):
                listener.exitAttribute(self)




    def attribute(self):

        localctx = PlantUMLParser.AttributeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_attribute)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 190
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 31457280) != 0):
                self.state = 189
                self.visibility()


            self.state = 192
            self.match(PlantUMLParser.ID)
            self.state = 193
            self.match(PlantUMLParser.T__6)
            self.state = 194
            self.dType()
            self.state = 195
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MethodContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.name = None # Token

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def visibility(self):
            return self.getTypedRuleContext(PlantUMLParser.VisibilityContext,0)


        def modifier(self):
            return self.getTypedRuleContext(PlantUMLParser.ModifierContext,0)


        def parameter(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.ParameterContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.ParameterContext,i)


        def dType(self):
            return self.getTypedRuleContext(PlantUMLParser.DTypeContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_method

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMethod" ):
                listener.enterMethod(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMethod" ):
                listener.exitMethod(self)




    def method(self):

        localctx = PlantUMLParser.MethodContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_method)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 198
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 31457280) != 0):
                self.state = 197
                self.visibility()


            self.state = 201
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==34 or _la==35:
                self.state = 200
                self.modifier()


            self.state = 203
            localctx.name = self.match(PlantUMLParser.ID)
            self.state = 204
            self.match(PlantUMLParser.T__15)
            self.state = 210
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==38:
                self.state = 205
                self.parameter()
                self.state = 208
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==17:
                    self.state = 206
                    self.match(PlantUMLParser.T__16)
                    self.state = 207
                    self.parameter()




            self.state = 212
            self.match(PlantUMLParser.T__17)
            self.state = 215
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==7:
                self.state = 213
                self.match(PlantUMLParser.T__6)
                self.state = 214
                self.dType()


            self.state = 217
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ParameterContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.name = None # Token

        def dType(self):
            return self.getTypedRuleContext(PlantUMLParser.DTypeContext,0)


        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def value(self):
            return self.getTypedRuleContext(PlantUMLParser.ValueContext,0)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_parameter

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterParameter" ):
                listener.enterParameter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitParameter" ):
                listener.exitParameter(self)




    def parameter(self):

        localctx = PlantUMLParser.ParameterContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_parameter)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 219
            localctx.name = self.match(PlantUMLParser.ID)
            self.state = 220
            self.match(PlantUMLParser.T__6)
            self.state = 221
            self.dType()
            self.state = 224
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==19:
                self.state = 222
                self.match(PlantUMLParser.T__18)
                self.state = 223
                self.value()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def INT(self):
            return self.getToken(PlantUMLParser.INT, 0)

        def FLOAT(self):
            return self.getToken(PlantUMLParser.FLOAT, 0)

        def D_QUOTE(self, i:int=None):
            if i is None:
                return self.getTokens(PlantUMLParser.D_QUOTE)
            else:
                return self.getToken(PlantUMLParser.D_QUOTE, i)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)




    def value(self):

        localctx = PlantUMLParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_value)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 227
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==44:
                self.state = 226
                self.match(PlantUMLParser.D_QUOTE)


            self.state = 229
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 1924145348608) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 231
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==44:
                self.state = 230
                self.match(PlantUMLParser.D_QUOTE)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DTypeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def primitiveData(self):
            return self.getTypedRuleContext(PlantUMLParser.PrimitiveDataContext,0)


        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_dType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDType" ):
                listener.enterDType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDType" ):
                listener.exitDType(self)




    def dType(self):

        localctx = PlantUMLParser.DTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_dType)
        try:
            self.state = 235
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [25, 26, 27, 28, 29, 30, 31, 32, 33]:
                self.enterOuterAlt(localctx, 1)
                self.state = 233
                self.primitiveData()
                pass
            elif token in [38]:
                self.enterOuterAlt(localctx, 2)
                self.state = 234
                self.match(PlantUMLParser.ID)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EnumerationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def enumLiteral(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PlantUMLParser.EnumLiteralContext)
            else:
                return self.getTypedRuleContext(PlantUMLParser.EnumLiteralContext,i)


        def getRuleIndex(self):
            return PlantUMLParser.RULE_enumeration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnumeration" ):
                listener.enterEnumeration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnumeration" ):
                listener.exitEnumeration(self)




    def enumeration(self):

        localctx = PlantUMLParser.EnumerationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_enumeration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 237
            self.match(PlantUMLParser.T__19)
            self.state = 238
            self.match(PlantUMLParser.ID)
            self.state = 239
            self.match(PlantUMLParser.T__3)
            self.state = 240
            self.match(PlantUMLParser.NL)
            self.state = 244
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==38:
                self.state = 241
                self.enumLiteral()
                self.state = 246
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 247
            self.match(PlantUMLParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EnumLiteralContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(PlantUMLParser.ID, 0)

        def NL(self):
            return self.getToken(PlantUMLParser.NL, 0)

        def getRuleIndex(self):
            return PlantUMLParser.RULE_enumLiteral

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnumLiteral" ):
                listener.enterEnumLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnumLiteral" ):
                listener.exitEnumLiteral(self)




    def enumLiteral(self):

        localctx = PlantUMLParser.EnumLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_enumLiteral)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 249
            self.match(PlantUMLParser.ID)
            self.state = 250
            self.match(PlantUMLParser.NL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VisibilityContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_visibility

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVisibility" ):
                listener.enterVisibility(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVisibility" ):
                listener.exitVisibility(self)




    def visibility(self):

        localctx = PlantUMLParser.VisibilityContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_visibility)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 252
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 31457280) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PrimitiveDataContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_primitiveData

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPrimitiveData" ):
                listener.enterPrimitiveData(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPrimitiveData" ):
                listener.exitPrimitiveData(self)




    def primitiveData(self):

        localctx = PlantUMLParser.PrimitiveDataContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_primitiveData)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 254
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 17146314752) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ModifierContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PlantUMLParser.RULE_modifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterModifier" ):
                listener.enterModifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitModifier" ):
                listener.exitModifier(self)




    def modifier(self):

        localctx = PlantUMLParser.ModifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_modifier)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 256
            _la = self._input.LA(1)
            if not(_la==34 or _la==35):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import os
from abc import ABC, abstractmethod
from besser.BUML.metamodel.structural import Model

# Interface for code generators
class GeneratorInterface(ABC):

    @abstractmethod
    def __init__(self, model: Model, output_dir: str = None):
        self.model = model
        self.output_dir = output_dir

    @abstractmethod
    def generate(self):
        pass

    @property
    def model(self) -> Model:
        return self.__model

    @model.setter
    def model(self, model: Model):
        self.__model = model

    @property
    def output_dir(self) -> str:
        return self.__output_dir

    @output_dir.setter
    def output_dir(self, output_dir: str):
        self.__output_dir = output_dir

    def build_generation_path(self, file_name:str) -> str:
        file_path = os.path.join(self.build_generation_dir(), file_name)
        return file_path

    def build_generation_dir(self) -> str:
        if self.output_dir is not None:
            os.makedirs(self.output_dir, exist_ok=True)
            file_path = os.path.join(self.output_dir)
        else:
            working_path = os.path.abspath('')
            os.makedirs(os.path.join(working_path, "output"), exist_ok=True)
            file_path = os.path.join(working_path, "output")
        return file_path



import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
import itertools

class PydanticGenerator(GeneratorInterface):
    """
    PydanticGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the Python domain model code based on the input B-UML model.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        backend (bool, optional): A boolean flag indicating whether the generator should generate code for a backend API.
        nested_creations (bool, optional): This parameter determines how entities are linked in the API request. 
                                            If set to True, both nested creations and linking by the ID of the entity 
                                            are enabled. If set to False, only the ID of the linked entity will be used.
                                            The default value is False.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """
    def __init__(self, model: DomainModel, backend: bool = False, nested_creations: bool = False, output_dir: str = None):
        super().__init__(model, output_dir)
        self.domain_model = model
        self.backend = backend
        self.nested_creations = nested_creations

    def generate(self):
        """
        Generates Python domain model code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named pydantic_classes.py 
        """
        file_path = self.build_generation_path(file_name="pydantic_classes.py")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(
            loader=FileSystemLoader(templates_path),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
            extensions=['jinja2.ext.do']
        )
        template = env.get_template('pydantic_classes_template.py.j2')
        with open(file_path, mode="w", newline='\n') as f:
            generated_code = template.render(
                domain=self.domain_model,
                backend=self.backend,
                nested_creations=self.nested_creations
            )
            f.write(generated_code)
            print("Code generated in the location: " + file_path)


import os
import configparser
import docker
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.generators.rest_api import RESTAPIGenerator
from besser.generators.sql_alchemy import SQLAlchemyGenerator
from besser.generators.pydantic_classes import PydanticGenerator
from besser.generators.backend.docker_files import generate_docker_files

class BackendGenerator(GeneratorInterface):
    """
    BackendGenerator is a class that implements the GeneratorInterface and is responsible for generating
    a Backend model code with a REST API using FAST API framework, SQLAlchemy and a Pydantic model based on the input B-UML model.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        http_methods (list, optional): A list of HTTP methods to be used in the REST API. Defaults to All.
        nested_creations (bool, optional): This parameter specifies how entities are linked in the API request. If set to True, the API expects 
                                identifiers and links entities based on these IDs. If set to False, the API handles the creation of 
                                new entities based on the data provided in the request. Defaults to True
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
        docker_image (bool, optional): Flag to indicate if Docker image generation is required. Defaults to False.
        docker_config_path (str, optional): The path to the docker configuration file to auto upload the image. Defaults to None.
    """

    def __init__(self, model: DomainModel, http_methods: list = None, nested_creations: bool = False, output_dir: str = None, docker_image: bool = False, docker_config_path: str = None):
        super().__init__(model, output_dir)
        allowed_methods = ["GET", "POST", "PUT", "DELETE"]
        if not http_methods:
            http_methods = allowed_methods
        else:
            http_methods = [method for method in http_methods if method in allowed_methods]
        self.http_methods = http_methods
        self.nested_creations = nested_creations
        self.docker_image = docker_image
        self.docker_config_path = docker_config_path
        self.config = self.load_config()

    def load_config(self):
        """
        Loads the configuration from the specified config file path if provided.

        Returns:
            dict: A dictionary containing the Docker configuration, or None if the config file path is not provided or file not found.
        """
        if self.docker_config_path and os.path.exists(self.docker_config_path):
            config = configparser.ConfigParser()
            config.read(self.docker_config_path)
            return {
                "docker_username": config.get("DEFAULT", "docker_username"),
                "docker_password": config.get("DEFAULT", "docker_password"),
                "docker_image_name": config.get("DEFAULT", "docker_image_name"),
                "docker_repository": config.get("DEFAULT", "docker_repository"),
                "docker_tag": config.get("DEFAULT", "docker_tag"),
                "docker_port": config.get("DEFAULT", "docker_port"),
            }
        else:
            if not self.docker_config_path:
                print("No configuration docker file.")
            else:
                print(f"Configuration file not found at path: {self.docker_config_path}")
            return None

    def generate(self):
        """
        Generates Backend model code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output_backend folder.

        Returns:
            None, but stores the generated code as files main_api.py, sql_alchemy.py and pydantic_classes.py
        """
        if self.output_dir is not None:
            backend_folder_path = self.output_dir
            os.makedirs(backend_folder_path, exist_ok=True)
            print(f"Backend folder created at {backend_folder_path}")
        else:
            backend_folder_path = os.path.join(os.path.abspath(''), "output_backend")
            os.makedirs(backend_folder_path, exist_ok=True)
            print(f"Backend folder created at {backend_folder_path}")

        docker_port = self.config["docker_port"] if self.config else 8000  # Use default port if config not provided

        rest_api = RESTAPIGenerator(model=self.model, http_methods=self.http_methods, nested_creations=self.nested_creations, output_dir=backend_folder_path, backend=True, port=docker_port)
        rest_api.generate()

        sql_alchemy = SQLAlchemyGenerator(model=self.model, output_dir=backend_folder_path)
        sql_alchemy.generate()

        pydantic_model = PydanticGenerator(model=self.model, output_dir=backend_folder_path, backend=True, nested_creations=self.nested_creations)
        pydantic_model.generate()

        if self.docker_image:
            if self.config:
                self.build_and_push_docker_image(backend_folder_path)
            else:
                generate_docker_files(backend_folder_path)

    def build_and_push_docker_image(self, backend_folder_path):
        """
        Builds and pushes the Docker image based on the provided backend folder path.

        Args:
            backend_folder_path (str): The path to the backend folder containing the generated code.

        Returns:
            None
        """
        docker_port = self.config["docker_port"]

        dockerfile_content = f"""
        FROM python:3.9-slim
        WORKDIR /app

        COPY main_api.py /app
        COPY pydantic_classes.py /app
        COPY sql_alchemy.py /app

        RUN pip install requests==2.31.0
        RUN pip install fastapi==0.110.0
        RUN pip install pydantic==2.6.3
        RUN pip install uvicorn==0.28.0
        RUN pip install SQLAlchemy==2.0.29
        RUN pip install httpx==0.27.0

        EXPOSE {docker_port}
        CMD ["python", "main_api.py"]
        """

        with open(os.path.join(backend_folder_path, 'Dockerfile'), 'w') as dockerfile:
            dockerfile.write(dockerfile_content)

        image_name = self.config["docker_image_name"]
        repository = self.config["docker_repository"]
        tag = self.config["docker_tag"]
        full_image_name = f"{repository}/{image_name}:{tag}"

        # Create docker client
        client = docker.from_env()

        # Create docker image
        image, build_logs = client.images.build(
            path=backend_folder_path,
            rm=True,
            tag=full_image_name
        )

        for log in build_logs:
            print(log)

        # Docker login
        username = self.config["docker_username"]
        password = self.config["docker_password"]
        client.login(username=username, password=password)

        # Push Docker image
        resp = client.api.push(full_image_name, stream=True, decode=True)
        for line in resp:
            print(line)



"""
BUML code for VGG16 from the literature.
"""
from besser.BUML.metamodel.nn import NN, Conv2D, PoolingLayer, \
    LinearLayer, DropoutLayer, FlattenLayer
from besser.generators.nn.pytorch.pytorch_code_generator import (
    PytorchGenerator
)
from besser.generators.nn.tf.tf_code_generator import TFGenerator



features: NN = NN(name="features")
features.add_layer(Conv2D(name="f1", actv_func="relu", in_channels=3,
                          out_channels=64, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f2", actv_func="relu", in_channels=64,
                          out_channels=64, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(PoolingLayer(name="f3", actv_func=None, pooling_type="max",
                                dimension="2D", kernel_dim=[2, 2],
                                stride_dim=[2, 2]))
features.add_layer(Conv2D(name="f4", actv_func="relu", in_channels=64,
                          out_channels=128, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f5", actv_func="relu", in_channels=128,
                          out_channels=128, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(PoolingLayer(name="f6", actv_func=None, pooling_type="max",
                                dimension="2D", kernel_dim=[2, 2],
                                stride_dim=[2, 2]))
features.add_layer(Conv2D(name="f7", actv_func="relu", in_channels=128,
                          out_channels=256, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f8", actv_func="relu", in_channels=256,
                          out_channels=256, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f9", actv_func="relu", in_channels=256,
                          out_channels=256, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(PoolingLayer(name="f10", actv_func=None, pooling_type="max",
                                dimension="2D", kernel_dim=[2, 2],
                                stride_dim=[2, 2]))
features.add_layer(Conv2D(name="f11", actv_func="relu", in_channels=256,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f12", actv_func="relu", in_channels=512,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f13", actv_func="relu", in_channels=512,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(PoolingLayer(name="f14", actv_func=None, pooling_type="max",
                                dimension="2D", kernel_dim=[2, 2],
                                stride_dim=[2, 2]))
features.add_layer(Conv2D(name="f15", actv_func="relu", in_channels=512,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f16", actv_func="relu", in_channels=512,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(Conv2D(name="f17", actv_func="relu", in_channels=512,
                          out_channels=512, kernel_dim=[3, 3],
                          stride_dim=[1, 1], padding_amount=1))
features.add_layer(PoolingLayer(name="f18", actv_func=None, pooling_type="max",
                                dimension="2D", kernel_dim=[2, 2],
                                stride_dim=[2, 2]))

classifier: NN = NN(name="classifier")
classifier.add_layer(LinearLayer(name="c1", actv_func="relu",
                                 in_features=25088, out_features=4096))
classifier.add_layer(DropoutLayer(name="c2", rate=0.5))
classifier.add_layer(LinearLayer(name="c3", actv_func="relu",
                                 in_features=4096, out_features=4096))
classifier.add_layer(DropoutLayer(name="c4", rate=0.5))
classifier.add_layer(LinearLayer(name="c5", actv_func=None,
                                 in_features=4096, out_features=1000))


vgg16: NN = NN(name="vgg16")
vgg16.add_sub_nn(features)
vgg16.add_layer(PoolingLayer(name="p1", actv_func=None,
                             pooling_type="adaptive_average", dimension="2D",
                             output_dim=[7, 7]))
vgg16.add_layer(FlattenLayer(name="f1", actv_func=None))
vgg16.add_sub_nn(classifier)

pytorch_model = PytorchGenerator(model=vgg16, output_dir="output/vgg16")
pytorch_model.generate()
tf_model = TFGenerator(model=vgg16, output_dir="output/vgg16")
tf_model.generate()



def generate_docker_files(path: str = "output_backend"):
    generate_dockerfile(path)
    generate_docker_image(path)
    pass

def generate_dockerfile(path: str):
    with open(path + '/Dockerfile', 'w') as dockerfile:
        dockerfile.write('''FROM python:3.9-slim
WORKDIR /app

COPY main_api.py /app
COPY pydantic_classes.py /app
COPY sql_alchemy.py /app

RUN pip install requests==2.31.0
RUN pip install fastapi==0.110.0
RUN pip install pydantic==2.6.3
RUN pip install uvicorn==0.28.0
RUN pip install SQLAlchemy==2.0.29
RUN pip install httpx==0.27.0

EXPOSE 8000
CMD ["python", "main_api.py"]
'''
        )

def generate_docker_image(path: str):
    with open(path + '/create_docker_image.py', 'w') as create_image:
        create_image.write('''import docker
import os
import argparse
import sys

def build_args_parser():
    parser = argparse.ArgumentParser(description='Generating and pushing the image to DockerHub')
    parser.add_argument('-u', '--username',
                        required=True,
                        help='docker username')
    parser.add_argument('-p', '--password',
                        required=True,
                        help='docker password')
    parser.add_argument('-i', '--image',
                        required=True,
                        help='docker image name')
    parser.add_argument('-t', '--tag',
                        required=True,
                        help='tag name')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    return parser

def main():
    args = build_args_parser().parse_args()
    # Get dockerfile path
    dockerfile_path = os.path.dirname(os.path.abspath(__file__))
    # Create docker client
    client = docker.from_env()
    # Create docker image
    tag = args.username + '/' + args.image + ':' + args.tag
    image, build_logs = client.images.build(
        path = dockerfile_path,
        rm = True,
        tag = tag
    )

    for log in build_logs:
        print(log)

    client.login(username=args.username, password=args.password)
    resp = client.api.push(
        tag,
        stream=True,
        decode=True,
    )
    
    for line in resp:
        print(line)

if __name__ == '__main__':
    main()
'''
        )



from besser.BUML.metamodel.structural import DomainModel, Class, Property, \
    Multiplicity, BinaryAssociation,Constraint,\
    StringType, IntegerType, DateType
from besser.BUML.metamodel.object import *
import datetime

#############################################
#   Library - structural model definition   #
#############################################

# Library attributes definition
library_name: Property = Property(name="name", type=StringType)
address: Property = Property(name="address", type=StringType)
# Library class definition
library: Class = Class (name="Library", attributes={library_name, address})

# Book attributes definition
title: Property = Property(name="title", type=StringType)
pages: Property = Property(name="pages", type=IntegerType)
release: Property = Property(name="release", type=DateType)
# Book class definition
book: Class = Class (name="Book", attributes={title, pages, release})

# Author attributes definition
author_name: Property = Property(name="name", type=StringType)
email: Property = Property(name="email", type=StringType)
# Author class definition
author: Class = Class (name="Author", attributes={author_name, email})

# Library-Book association definition
located_in: Property = Property(name="locatedIn",type=library, multiplicity=Multiplicity(1, 1))
has: Property = Property(name="has", type=book, multiplicity=Multiplicity(0, "*"))
lib_book_association: BinaryAssociation = BinaryAssociation(name="lib_book_assoc", ends={located_in, has})

# Book-Author association definition
publishes: Property = Property(name="publishes", type=book, multiplicity=Multiplicity(0, "*"))
written_by: Property = Property(name="writtenBy", type=author, multiplicity=Multiplicity(1, "*"))
book_author_association: BinaryAssociation = BinaryAssociation(name="book_author_assoc", ends={written_by, publishes})

constraintPageNumber: Constraint = Constraint(name = "libraryPageNumber",context=library,expression="context Library inv inv1: self.has ->forAll(b:Book|b.pages>0)",language="OCL")

constraintBookPageNumber: Constraint = Constraint(name = "BookPageNumber",context=book,expression="context Book inv inv2: self.pages>0",language="OCL")

constraintTitleIncorrect: Constraint = Constraint(name = "BookTitleInc",context=book,expression="context Book inv inv2: self.title= 'NI')",language="OCL")

constraintTitleCorrect: Constraint = Constraint(name = "BookTitle",context=book,expression="context Book inv inv2: self.title <> 'NI'",language="OCL")

constraintLibraryExists: Constraint = Constraint(name = "LibarayExistsConst",context=library,expression="context Library inv inv3: self.has->exists( i_book : Book | i_book.pages <= 110 )",language="OCL")

# constraintLibraryExclude: Constraint = Constraint(name = "LibarayExclude",context=library,expression="context Library inv inv3: self.has->excludes()",language="OCL")
constraintLibrarySize: Constraint = Constraint(name = "LibaraySize",context=library,expression="context Library inv inv3: self.has->size()>1",language="OCL")

constraintLibraryCollect: Constraint = Constraint(name = "LibaryCollect",context=library,
expression="context Library inv inv3: self.has->collect(i_book : Book | i_book.pages <= 110)"
           "->size()>0",language="OCL")

constraintLibraryIf: Constraint = Constraint(name ="LibaryIf", context=library,
expression="context Library inv inv3: if self.name <> 'NI' then self.has->exists"
           "( i_book : Book | i_book.pages <= 110 ) else self.has->forAll(b:Book|b.pages>0)"
           " endif", language="OCL")

constraintLibraryElse: Constraint = Constraint(name ="LibaryElse", context=library,
expression="context Library inv inv3: if self.name = 'NI' then self.has->exists"
           "( i_book : Book | i_book.pages <= 110 ) else self.has->forAll(b:Book|b.pages>0)"
           " endif", language="OCL")

constraintLibraryElseFalse: Constraint = Constraint(name ="LibaryElseFalse", context=library, expression="context Library inv inv3: if self.name = 'NI' then self.has->exists( i_book : Book | i_book.pages <= 110 ) else self.has->forAll(b:Book|b.pages<0) endif", language="OCL")


constraintNameOCLIsTypeOf: Constraint = Constraint(name ="BookIsTypeOfStr", context=book, expression="context Book inv inv3: self.title.oclIsTypeOf(String)", language="OCL")

constraintNameOCLIsTypeOfInt: Constraint = Constraint(name ="BookIsTypeOfInt", context=book, expression="context Book inv inv3: self.pages.oclIsTypeOf(Integer)", language="OCL")

constraintNameOCLIsTypeOfIntFalse: Constraint = Constraint(name ="BookIsTypeOfIntFalse", context=book, expression="context Book inv inv3: self.pages.oclIsTypeOf(String)", language="OCL")
constraintifElseSize: Constraint = Constraint(name ="constraintifElseSize", context=library, expression="context Library inv inv3: if self.name <> 'NI' then self.has->exists( i_book : Book | i_book.pages <= 110 )->size()<3 else self.has->forAll(b:Book|b.pages>0) endif", language="OCL")


# Domain model definition
library_model : DomainModel = DomainModel(name="Library_model", types={library, book, author}, 
                                          associations={lib_book_association, book_author_association},
                                          constraints={
                                              constraintBookPageNumber,
                                                       # constraintTitleIncorrect,
                                                       constraintPageNumber,
                                                       constraintTitleCorrect,
                                                       constraintLibraryExists,
                                                       constraintLibrarySize,
                                                       constraintLibraryCollect,
                                                       constraintLibraryIf,
                                                       constraintLibraryElse,
                                                       constraintLibraryElseFalse,
                                                       constraintNameOCLIsTypeOf,
                                                       constraintNameOCLIsTypeOfInt,
                                                       constraintNameOCLIsTypeOfIntFalse
                                          }

                                            # constraints={constraintLibrarySize}
                                          )


#########################################
#   Library - object model definition   #
#########################################


# Library object attributes
library_obj_name: AttributeLink = AttributeLink(attribute=library_name, value=DataValue(classifier=StringType, value="Library test"))
library_obj_address: AttributeLink = AttributeLink(attribute=address, value=DataValue(classifier=StringType, value="street 123"))
# Library object
library_obj: Object = Object(name="Library_Object", classifier=library, slots=[library_obj_name, library_obj_address])

# Book object attributes
book_obj_name: AttributeLink = AttributeLink(attribute=title, value=DataValue(classifier=StringType, value="Book tittle"))
book_obj_pages: AttributeLink = AttributeLink(attribute=pages, value=DataValue(classifier=IntegerType, value=100))
book_obj_release: AttributeLink = AttributeLink(attribute=release, value=DataValue(classifier=DateType, value=datetime.datetime(2020, 3, 15)))
# Book object
book_obj: Object = Object(name="Book_Object", classifier=book, slots=[book_obj_name, book_obj_pages])

# Book_2 object attributes

book_obj_name_2: AttributeLink = AttributeLink(attribute=title, value=DataValue(classifier=StringType, value="Book tittle_2"))
book_obj_pages_2: AttributeLink = AttributeLink(attribute=pages, value=DataValue(classifier=IntegerType, value=400))
book_obj_release_2: AttributeLink = AttributeLink(attribute=release, value=DataValue(classifier=DateType, value=datetime.datetime(2024, 3, 15)))
# Book object
book_obj_2: Object = Object(name="Book_2_Object", classifier=book, slots=[book_obj_name_2, book_obj_pages_2])

# Author object attributes
author_obj_name: AttributeLink = AttributeLink(attribute=author_name, value=DataValue(classifier=StringType, value="John Doe"))
author_obj_email: AttributeLink = AttributeLink(attribute=email, value=DataValue(classifier=StringType, value="john@doe.com"))
# Author object
author_obj: Object = Object(name="Author_Object", classifier=author, slots=[author_obj_name, author_obj_email])

# Book object and Author object link
book_link_end1: LinkEnd = LinkEnd(name="book_end1", association_end=publishes, object=book_obj)
author_link_end: LinkEnd = LinkEnd(name="author_end", association_end=written_by, object=author_obj)
author_book_link: Link = Link(name="author_book_link", association=book_author_association, connections=[book_link_end1,author_link_end])

# Book Library and Book object link
book_link_end2: LinkEnd = LinkEnd(name="book_end2", association_end=has, object=book_obj)
library_link_end: LinkEnd = LinkEnd(name="library_end", association_end=located_in, object=library_obj)
library_book_link: Link = Link(name="library_book_link", association=book_author_association, connections=[book_link_end2,library_link_end])

# Book object and Author object link
book_link_end2: LinkEnd = LinkEnd(name="book_end3", association_end=publishes, object=book_obj_2)
author_link_end2: LinkEnd = LinkEnd(name="author_end2", association_end=written_by, object=author_obj)
author_book_link2: Link = Link(name="author_book_link2", association=book_author_association, connections=[book_link_end2,author_link_end2])

# Book Library and Book object link
book_link_end3: LinkEnd = LinkEnd(name="book_end4", association_end=has, object=book_obj_2)
library_link_end3: LinkEnd = LinkEnd(name="library_end3", association_end=located_in, object=library_obj)
library_book_link3: Link = Link(name="library_book_link3", association=book_author_association, connections=[book_link_end3,library_link_end3])



# Object model definition
object_model: ObjectModel = ObjectModel(name="Object_model", instances={library_obj, author_obj, book_obj,book_obj_2}, links={author_book_link, library_book_link,author_book_link2,library_book_link3})


import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface


class RDFGenerator(GeneratorInterface):
    """
    RDFGenerator is a class that implements the GeneratorInterface and produces the vocabulary spec in RDF Turtle format 
    for a structural model defined with B-UML.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    TYPES = {
        "int": "integer",
        "str": "string",
        "float": "float",
        "bool": "boolean",
        "time": "time",
        "date": "date",
        "datetime": "dateTime",
        "timedelta": "duration",
    }

    def __init__(self, model: DomainModel, output_dir: str = None):
        super().__init__(model, output_dir)

    def generate(self):
        """
        Generates RDF vocabulary on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named vocabulary.ttl
        """
        file_path = self.build_generation_path(file_name="vocabulary.ttl")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path), trim_blocks=True, lstrip_blocks=True)
        template = env.get_template('rdf_template.j2')
        with open(file_path, mode="w") as f:
            generated_code = template.render(model=self.model, types=self.TYPES)
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



"""
This module provides the `SetupLayerSyntax` class and the 
`get_tensorop_syntax` function.
The `SetupLayerSyntax` class is used to define the syntax 
of layers in TensorFlow, while `get_tensorop_syntax` defines the 
tensorOps.
"""

from besser.generators.nn import utils_nn as utils


class SetupLayerSyntax:
    """
    This class is used to map the BUML synatx for layers to TensorFlow
    syntax.
    It processes the layers based on their type.
    """
    def __init__(self, layer, modules_details):
        self.layer = layer
        self.modules_details: dict = modules_details
        self.permute_out: bool = None
        self.permute_in: bool = None

    def setup_general_layer(self):
        """It defines the syntax of general layers."""
        actv_func = self.setup_actv_func()
        cls_name = self.layer.__class__.__name__
        lyr_name = self.layer.name
        lyr = f"self.{lyr_name} = layers"
        if cls_name == "LinearLayer":
            lyr = (
                f"{lyr}.Dense(units={self.layer.out_features}, "
                f"activation={actv_func})"
            )
        elif cls_name == "FlattenLayer":
            lyr = f"{lyr}.Flatten()"
        else: #cls_name == "EmbeddingLayer"
            lyr = (
                f"{lyr}.Embedding(input_dim={self.layer.num_embeddings}, "
                f"output_dim={self.layer.embedding_dim})"
            )
        return lyr

    def setup_layer_modifier(self):
        """It defines the syntax of layers' modifiers."""
        cls_name = self.layer.__class__.__name__
        parent_cls = self.layer.__class__.mro()[1].__name__
        lyr_name = self.layer.name
        lyr = f"self.{lyr_name} = layers"
        if parent_cls == "NormalizationLayer":
            if cls_name == "BatchNormLayer":
                lyr = f"{lyr}.BatchNormalization()"
            else: #cls_name == "LayerNormLayer"
                norm_shape = self.layer.normalized_shape
                if norm_shape[1] is not None:
                    if norm_shape[2] is not None:
                        shape = [norm_shape[0], norm_shape[1], norm_shape[2]]
                    else:
                        shape = [norm_shape[0], norm_shape[1]]
                else:
                    shape = [norm_shape[0]]

                lyr = f"{lyr}.LayerNormalization(axis={shape})"
        else: #cls_name == "DropoutLayer"
            lyr = f"{lyr}.Dropout(rate={self.layer.rate})"
        return lyr

    def setup_rnn(self):
        """It defines the syntax of rnn layers."""
        actv_func = self.setup_actv_func()
        cls_name = self.layer.__class__.__name__
        lyr_name = self.layer.name
        layer_type = cls_name[:-5]
        lyr = (
            f"layers.{layer_type}(units={self.layer.hidden_size}, "
            f"activation={actv_func}, dropout={self.layer.dropout}"
        )

        if self.layer.return_type == "full":
            lyr = f"{lyr}, return_sequences=True)"
        elif self.layer.return_type == "hidden":
            lyr = f"{lyr}, return_state=True)"
        else:
            lyr = f"{lyr})"

        if self.layer.bidirectional is True:
            lyr = f"self.{lyr_name} = layers.Bidirectional({lyr})"
        else:
            lyr = f"self.{lyr_name} = {lyr}"

        #self.modules_details is returned to have the same structure
        #of setup_rnn from torch
        return lyr, self.modules_details

    def setup_actv_func(self):
        """
        It formats the activation function as attribute of the layer.
        """
        if hasattr(self.layer, 'actv_func'):
            activ = self.layer.actv_func
            list_func = ["relu", "tanh", "sigmoid","softmax", "leaky_relu"]
            if activ is not None:
                if activ in list_func:
                    return f"'{self.layer.actv_func}'"
                else:
                    return f"{self.layer.actv_func}"
            else:
                return None

    def setup_conv(self, lyr_name, cls_name):
        """It defines the syntax of convolutional layers."""
        actv_func = self.setup_actv_func()
        dim = cls_name[-2:-1]
        filters = self.layer.out_channels
        pad_type = self.layer.padding_type
        kernel = utils.format_value(self.layer.kernel_dim)
        stride = utils.format_value(self.layer.stride_dim)
        pad_amount = self.layer.padding_amount
        lyr = ""
        if pad_amount != 0:
            lyr = (
                f"self.{lyr_name}_pad = layers.ZeroPadding{dim}D("
                f"padding={pad_amount})#"
            )
        lyr = (
            f"{lyr}self.{lyr_name} = layers.Conv{dim}D(filters={filters}, "
            f"kernel_size={kernel}, strides={stride}, "
            f"padding='{pad_type}', activation={actv_func})"
        )
        return lyr

    def setup_pooling(self, lyr_name):
        """It defines the syntax of pooling layers."""
        pl_type = self.layer.pooling_type
        dim = self.layer.dimension[-2:-1]
        if pl_type == "max" or pl_type == "average":
            pl = "MaxPool" if pl_type == "max" else "AveragePooling"
            kernel = utils.format_value(self.layer.kernel_dim)
            stride = utils.format_value(self.layer.stride_dim)
            pad_type = self.layer.padding_type
            lyr = (
                f"self.{lyr_name} = layers.{pl}{dim}D(pool_size={kernel}, "
                f"strides={stride}, padding='{pad_type}')"
            )
        elif pl_type.startswith("global"):
            typ = pl_type.split("_")[1]
            pl = f"Global{typ[0].upper()}{typ[1:]}Pooling"
            lyr = (
                f"self.{lyr_name} = layers.{pl}{dim}D()"
            )
        else:
            if pl_type == "adaptive_average":
                pl = "AdaptiveAveragePooling"
            else:
                pl = "AdaptiveMaxPooling"

            size = utils.format_value(self.layer.output_dim)
            lyr = (
                f"self.{lyr_name} = tfa.layers.{pl}{dim}D(output_size={size})"
            )
        return lyr

    def setup_cnn(self):
        """It defines the syntax of cnn layers (conv and pooling)."""
        cls_name = self.layer.__class__.__name__
        lyr_name = self.layer.name

        if cls_name == "PoolingLayer":
            lyr = self.setup_pooling(lyr_name)
        else:
            lyr = self.setup_conv(lyr_name, cls_name)

        #self.modules_details is returned to have the same structure
        #of setup_rnn from torch
        return lyr, self.modules_details


def get_tensorop_syntax(tensorop, modules_details, in_var=None):
    """It defines the syntax of tensorops."""
    prev_out_var, params = utils.get_tensorop_params(tensorop,
                                                     modules_details)
    if in_var is not None:
        prev_out_var = in_var

    tns_type = tensorop.tns_type
    if tns_type == "reshape":
        ts_op_synt = f"tf.reshape({prev_out_var}, {params})"
    elif tns_type == "concatenate":
        axis = tensorop.concatenate_dim
        ts_op_synt = f"tf.concat([{params}], axis={axis})"
    elif tns_type == "transpose":
        ts_op_synt = f"tf.transpose({prev_out_var}, perm=[{params}])"
    elif tns_type == "permute":
        ts_op_synt = f"tf.transpose({prev_out_var}, perm=[{params}])"
    elif tns_type == "multiply":
        ts_op_synt = f"tf.math.multiply({params})"
    else:
        ts_op_synt = f"tf.matmul({params})"
    return ts_op_synt



from besser.BUML.metamodel.object import *
from besser.BUML.metamodel.structural import *

########################
#   Structural model   #
########################

attribute1: Property = Property(name="A1", type=PrimitiveDataType("int"))
attribute2: Property = Property(name="A2", type=PrimitiveDataType("str"))
class1: Class = Class(name="class1", attributes={attribute1, attribute2})
class2: Class = Class(name="class2", attributes=set())
aend1: Property = Property(name="end1", type=class1, multiplicity=Multiplicity(0, 1))
aend2: Property = Property(name="end2", type=class2, multiplicity=Multiplicity(0, "*"))
association1: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})

##########################
#   Object model tests   #
##########################

# Testing model initialization
def test_model_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    model: ObjectModel = ObjectModel(name="mymodel", instances={obj1,obj2}, links={})
    assert len(model.instances) == 2
    model_empty: ObjectModel = ObjectModel(name="mymodel", instances={}, links={})
    assert len(model_empty.instances) == 0

# Testing object initialization
def test_object_initialization():
    attr1: AttributeLink = AttributeLink(value=DataValue(classifier=PrimitiveDataType("int"), value=100), attribute=attribute1)
    obj1: Object = Object(name="object1", classifier=class1, slots=[attr1])
    assert len(obj1.slots) == 1
    assert obj1.classifier.name == "class1"

# Testing attribute link
def test_attributeLink_initialization():
    attr1: AttributeLink = AttributeLink(value=DataValue(classifier=PrimitiveDataType("int"), value=100), attribute=attribute1)
    attr2: AttributeLink = AttributeLink(value=DataValue(classifier=PrimitiveDataType("str"), value="test value"), attribute=attribute2)
    assert attr1.value.value == 100
    assert attr2.value.value == "test value"

# Testing Link inicialization
def test_link_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    link1: Link = Link(name="link1", association=association1, connections=[link_end1, link_end2])
    assert len(link1.connections) == 2
    assert link1.association.name == "association1"
    assert len(link1.association.ends) == 2

# Testing Link inicialization
def test_linkEnd_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    assert link_end1.association_end.name == "end1"
    assert link_end2.object.name == "object2"

# Testing link_ends() method
def test_link_ends_method():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    link1: Link = Link(name="link1", association=association1, connections=[link_end1, link_end2])
    assert len(obj1.link_ends()) == 1
    for link_end in obj1.link_ends():
        assert link_end.object.name == "object2"
    for link_end in obj2.link_ends():
        assert link_end.object.name == "object1"


import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.utilities import sort_by_timestamp

class SQLAlchemyGenerator(GeneratorInterface):
    """
    SQLAlchemyGenerator is a class that implements the GeneratorInterface and is responsible for generating
    SQLAlchemy code based on B-UML models.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    TYPES = {
        "int": "Integer",
        "str": "String(100)",
        "float": "Float",
        "bool": "Boolean",
        "time": "Time",
        "date": "Date",
        "datetime": "DateTime",
    }

    VALID_DBMS = {"sqlite", "postgresql", "mysql", "mssql", "mariadb"}

    def __init__(self, model: DomainModel, output_dir: str = None):
        super().__init__(model, output_dir)
        # Add enums to TYPES dictionary
        for enum in model.get_enumerations():
            self.TYPES[enum.name] = f"Enum('{enum.name}')"

    def get_ids(self):
        """
        Returns a dictionary with the class names as keys and the id attributes as values.
        """
        ids_dict = {}
        for cls in self.model.get_classes():
            # First, look for the first attribute marked with is_id
            id_attr = next((attr.name for attr in cls.attributes if attr.is_id), None)

            # If no attribute with is_id is found, check for an attribute with name "id"
            if not id_attr:
                id_attr = next((attr.name for attr in cls.attributes if attr.name == "id"), None)

            # Only add to the dictionary if an is_id or "id" attribute is found
            if id_attr:
                ids_dict[cls.name] = id_attr

        return ids_dict

    def get_foreign_keys(self):
        fkeys = dict()

        for association in self.model.associations:
            ends = list(association.ends)  # Convert set to list

            # One-to-one
            if ends[0].multiplicity.max == 1 and ends[1].multiplicity.max == 1:
                fkeys[association.name] = ends[0].type.name
                if ends[1].multiplicity.min == 0:
                    fkeys[association.name] = ends[1].type.name

            # Many to one
            elif ends[0].multiplicity.max > 1 and ends[1].multiplicity.max <= 1:
                fkeys[association.name] = ends[0].type.name

            elif ends[0].multiplicity.max <= 1 and ends[1].multiplicity.max > 1:
                fkeys[association.name] = ends[1].type.name

        return fkeys

    def generate(self, dbms: str = "sqlite"):
        """
        Generates SQLAlchemy code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Args:
            dbms (str, optional): The database management system to be used. Values allowed: "sqlite", "postgresql", "mysql",
            "mssql", or "mariadb". Defaults to "sqlite".

        Returns:
            None, but stores the generated code as a file named sql_alchemy.py 
        """
        if dbms not in self.VALID_DBMS:
            raise ValueError(f"Invalid DBMS. Valid options are {', '.join(self.VALID_DBMS)}.")

        file_path = self.build_generation_path(file_name="sql_alchemy.py")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('sql_alchemy_template.py.j2')
        with open(file_path, mode="w", encoding="utf-8") as f:
            generated_code = template.render(
                classes=self.model.classes_sorted_by_inheritance(),
                types=self.TYPES,
                associations=self.model.associations,
                enumerations=self.model.get_enumerations(),
                model_name=self.model.name,
                dbms=dbms,
                ids=self.get_ids(),
                fkeys=self.get_foreign_keys(),
                sort=sort_by_timestamp
            )
            f.write(generated_code)
            print("Code generated successfully!")



import pytest
from besser.BUML.metamodel.structural import *
from besser.utilities import sort_by_timestamp

def test_named_element():
    named_element: NamedElement = NamedElement(name="element1")
    assert named_element.name == "element1"
    assert named_element.is_derived == False  # Default value should be False
    
    # Test setting is_derived
    named_element.is_derived = True
    assert named_element.is_derived == True
    
    # Test initialization with is_derived
    derived_element: NamedElement = NamedElement(name="element2", is_derived=True)
    assert derived_element.is_derived == True


def test_model_initialization():
    class1: Type = Type(name="element1")
    class2: Type = Type(name="element2")
    model: DomainModel = DomainModel(name="mymodel", types={class1, class2}, associations = None, packages = None, constraints = None)
    assert class1 in model.types
    assert class2 in model.types
    model_empty: DomainModel = DomainModel(name="mymodel", types = None, associations = None, packages = None, constraints = None)
    assert class1 not in model_empty.types
    assert class2 not in model_empty.types

# Testing the WFR for duplicate names in a model
def test_model_duplicated_names():
    with pytest.raises(ValueError) as excinfo:
        class1: Type = Type(name="name1")
        class2: Type = Type(name="name1")
        model: DomainModel = DomainModel(name="mymodel", types={class1, class2}, associations = None, packages = None, constraints = None)
    assert "The model cannot have types with duplicate names: name1" in str(excinfo.value)


# Testing attributes initialization
def test_attribute_initialization():
    class1: Type = Type(name="name1")
    attribute1: Property = Property(name="attribute1", owner = class1, type=PrimitiveDataType("int"),
                                    multiplicity=Multiplicity(0, 1))
    # assert attributes has proper type and multiplicity
    assert attribute1.type.name == "int"
    assert attribute1.multiplicity.min == 0
    assert attribute1.multiplicity.max == 1


# Testing attribute multiplicity and data type violations
def test_attribute_type_and_multiplicity_violation():
    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(0, -1))
    assert "Invalid max multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(-1, 1))
    assert "Invalid min multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(2, 1))
    assert "Invalid max multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"),
                                        multiplicity=Multiplicity(0, 0))
    assert "Invalid max multiplicity" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("invented_type"),
                                        multiplicity=Multiplicity(2, 1))
    assert "Invalid primitive data type" in str(excinfo.value)


# Testing class initialization
def test_class_initialization():
    class1: Class = Class(name="classA", attributes=set())
    class2: Class
    attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    attribute2: Property = Property(name="attribute2", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    reference1: Property = Property(name="reference1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    class2: Class = Class(name="classB", attributes={attribute1, attribute2})
    assert len(class2.attributes) == 2
    print(class2)


# Testing no duplicated names in class attributes
def test_duplicated_name_class():
    with pytest.raises(ValueError) as excinfo:
        class1 : Class
        attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
        attribute2: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
        class1 = Class(name="name1", attributes={attribute1, attribute2})
    assert "A class cannot have attributes with duplicate names: attribute1" in str(excinfo.value)

# Testing for no more than one id attribute in class
def test_more_than_one_id_class():
    with pytest.raises(ValueError) as excinfo:
        class1 : Class
        attribute1: Property = Property(name="attribute1", type=PrimitiveDataType("int"), is_id=True)
        attribute2: Property = Property(name="attribute2", type=PrimitiveDataType("int"), is_id=True)
        class1 = Class(name="name1", attributes={attribute1, attribute2})
    assert "A class cannot have more than one attribute marked as 'id'" in str(excinfo.value)

def test_association_initialization():
    class1: Class = Class(name="name1", attributes=set())
    class2: Class = Class(name="name2", attributes=set())
    aend1: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    aend2: Property = Property(name="end2", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
    association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})
    assert len(association.ends) == 2
    assert aend1 in association.ends
    assert aend1.owner == association
    assert class1.associations == {association}
    assert class1.association_ends() == {aend2}

# Testing the creation of a binary association cannot have more than two ends
def test_binary_association():
    with pytest.raises(ValueError) as excinfo:
        class1: Type = Type(name="name1")
        aend: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
        association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend})
    assert "A binary association must have exactly two ends" in str(excinfo.value)


# Testing the creation of an association class with an attribute
def test_association_class():
    class1: Class = Class(name="name1", attributes=None)
    class2: Class = Class(name="name2", attributes=None)
    aend1: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
    aend2: Property = Property(name="end2", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
    association: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})
    attribute1: Property = Property(name="attribute1", owner=None, type=PrimitiveDataType("int"), multiplicity=Multiplicity(0, 1))
    association_class: AssociationClass = AssociationClass(name="association_class1", attributes={attribute1}, association=association)
    assert len(association_class.attributes) == 1
    assert attribute1 in association_class.attributes
    assert association_class.association.name == "association1"


def test_generalization_initialization():
    attribute1: Property = Property(name="attribute1", owner = None, type=PrimitiveDataType("int"),
                                    multiplicity=Multiplicity(0, 1))
    class1: Class = Class(name="name1", attributes={attribute1})
    class2: Class = Class(name="name2", attributes=None)
    generalization: Generalization = Generalization(general=class1, specific=class2)
    assert generalization.general == class1
    assert generalization.specific == class2
    assert class2.generalizations == {generalization}
    assert class1.specializations() == {class2}
    assert class2.parents() == {class1}
    assert class2.specializations() == set()
    assert class2.all_attributes() == {attribute1}


def test_no_generalization_loop():
    with pytest.raises(ValueError) as excinfo:
        class1: Class = Class(name="name1", attributes=None)
        generalization: Generalization = Generalization(general=class1, specific=class1)
    assert "A class cannot be a generalization of itself" in str(excinfo.value)

def test_generalization_set_initialization():
    class1: Class = Class(name="name1", attributes=None)
    class2: Class = Class(name="name2", attributes=None)
    class3: Class = Class(name="name3", attributes=None)
    generalization1: Generalization = Generalization(general=class1, specific=class2)
    generalization2: Generalization = Generalization(general=class1, specific=class3)
    generalization_set: GeneralizationSet = GeneralizationSet(name="GeneralizationSet", generalizations={
        generalization1,generalization2}, is_disjoint=True, is_complete=True)
    assert generalization_set.is_disjoint == True
    assert generalization_set.is_complete == True
    assert class1.generalizations == {generalization1, generalization2}
    assert class1.specializations() == {class3, class2}
    assert class2.parents() == {class1}
    assert class3.parents() == {class1}
    assert class2.specializations() == set()
    
# Testing enumeration and literals initialization
def test_enumeration_initialization():
    literal1: EnumerationLiteral = EnumerationLiteral(name="literal_1", owner=None)
    literal2: EnumerationLiteral = EnumerationLiteral(name="literal_2", owner=None)
    literal3: EnumerationLiteral = EnumerationLiteral(name="literal_3", owner=None)
    enum: Enumeration = Enumeration(name="Enumeration", literals={literal1,literal2,literal3})
    assert len(enum.literals) == 3
    assert literal1.owner == enum
    
# Testing no duplicated literal names in an enumeration
def test_duplicated_name_literal():
    with pytest.raises(ValueError) as excinfo:
        literal1: EnumerationLiteral = EnumerationLiteral(name="duplicated_name", owner=None)
        literal2: EnumerationLiteral = EnumerationLiteral(name="duplicated_name", owner=None)
        enum1: Enumeration = Enumeration(name="Enumeration", literals={literal1,literal2})
    assert "An enumeration cannot have two literals with the same name" in str(excinfo.value)

# Testing method and parameters initialization
def test_method_initialization():
    parameter1: Parameter = Parameter(name="parameter_1", type=PrimitiveDataType(name="str"))
    parameter2: Parameter = Parameter(name="parameter_2", type=PrimitiveDataType(name="int"))
    method: Method = Method(name='method_1', is_abstract=True, parameters={parameter1, parameter2})
    cls: Class = Class(name="class_1", attributes={}, methods={method})
    method.type = cls
    assert len(method.parameters) == 2
    assert method.owner == cls
    assert method.name == "method_1"
    assert method.type.name == "class_1"

# Testing parameters with repeated name
def test_parameters_same_name():
    with pytest.raises(ValueError) as excinfo:
        parameter1: Parameter = Parameter(name="parameter_1", type=PrimitiveDataType(name="str"))
        parameter2: Parameter = Parameter(name="parameter_1", type=PrimitiveDataType(name="int"))
        method: Method = Method(name='method_1', is_abstract=True, parameters={parameter1, parameter2})
    assert "A method cannot have parameters with duplicate names: parameter_1" in str(excinfo.value)

# Testing sort attributes by timestamp
def test_sort_attributes():
    attribute1: Property = Property(name="attribute_1", type=PrimitiveDataType(name="str"))
    attribute2: Property = Property(name="attribute_2", type=PrimitiveDataType(name="int"))
    attribute3: Property = Property(name="attribute_3", type=PrimitiveDataType(name="int"))
    cls: Class = Class(name="class", attributes={attribute1, attribute2, attribute3})
    attributes = sort_by_timestamp(cls.attributes)
    assert len(attributes) == 3
    assert type(attributes) == list
    assert attributes[0] == attribute1
    assert attributes[1] == attribute2
    assert attributes[2] == attribute3

# Testing the classes_sorted_by_inheritance method
def test_classes_sorted_by_inheritance():
    cl1 = Class(name="c1")
    cl2 = Class(name="c2")
    cl3 = Class(name="c3")
    cl4 = Class(name="c4")
    cl5 = Class(name="c5")
    cl6 = Class(name="c6")
    cl7 = Class(name="c7")
    h1 = Generalization(general=cl7, specific=cl6)
    h2 = Generalization(general=cl6, specific=cl5)
    h3 = Generalization(general=cl5, specific=cl4)
    h4 = Generalization(general=cl4, specific=cl2)
    h5 = Generalization(general=cl4, specific=cl3)
    h6 = Generalization(general=cl3, specific=cl2)
    h7 = Generalization(general=cl2, specific=cl1)
    model = DomainModel(name="model", types={cl1,cl2,cl3,cl4,cl5,cl6,cl7}, generalizations={h1,h2,h3,h4,h5,h6,h7})
    classes = model.classes_sorted_by_inheritance()
    assert len(classes) == 7
    assert classes[0] == cl7
    assert classes[2] == cl5
    assert classes[4] == cl3
    assert classes[6] == cl1

# Testing synonyms of a Named Element
def test_synonyms():
    class_a: Class = Class(name="Library", synonyms=["synonym1", "synonym2", "synonym3"])
    assert len(class_a.synonyms) == 3
    assert class_a.synonyms[0] == "synonym1"
    assert class_a.synonyms[1] == "synonym2"
    assert class_a.synonyms[2] == "synonym3"

# Testing all_parents and inherited_attributes methods
# GrandParent (attr1)
#           /           \
#    Parent1 (attr2)    Parent2 (attr3)
#          |
#    Child (attr4)
#~The test verifies:
# The all_parents() method by checking that each class correctly identifies its parent classes up the hierarchy
# The inherited_attributes() method by verifying that each class correctly inherits attributes from its parent classes
# The all_attributes() method by checking that each class has both its own attributes and inherited attributes
# The test checks multiple inheritance paths and ensures that attributes are correctly inherited through the class hierarchy.

def test_all_parents_and_inherited_attributes():
    # Create a hierarchy of classes with attributes
    attribute1: Property = Property(name="attr1", type=PrimitiveDataType("int"), multiplicity=Multiplicity(1, 1))
    attribute2: Property = Property(name="attr2", type=PrimitiveDataType("str"), multiplicity=Multiplicity(1, 1))
    attribute3: Property = Property(name="attr3", type=PrimitiveDataType("bool"), multiplicity=Multiplicity(1, 1))
    attribute4: Property = Property(name="attr4", type=PrimitiveDataType("float"), multiplicity=Multiplicity(1, 1))
    
    # Create classes with their attributes
    class_grandparent: Class = Class(name="GrandParent", attributes={attribute1})
    class_parent1: Class = Class(name="Parent1", attributes={attribute2})
    class_parent2: Class = Class(name="Parent2", attributes={attribute3})
    class_child: Class = Class(name="Child", attributes={attribute4})
    
    # Create generalizations
    generalization1: Generalization = Generalization(general=class_grandparent, specific=class_parent1)
    generalization2: Generalization = Generalization(general=class_grandparent, specific=class_parent2)
    generalization3: Generalization = Generalization(general=class_parent1, specific=class_child)
    
    # Test all_parents()
    assert class_child.parents() == {class_parent1}
    assert class_child.all_parents() == {class_parent1, class_grandparent}
    assert class_parent1.all_parents() == {class_grandparent}
    assert class_parent2.all_parents() == {class_grandparent}
    assert class_grandparent.all_parents() == set()
    
    # Test inherited_attributes()
    assert class_child.inherited_attributes() == {attribute1, attribute2}
    assert class_parent1.inherited_attributes() == {attribute1}
    assert class_parent2.inherited_attributes() == {attribute1}
    assert class_grandparent.inherited_attributes() == set()
    
    # Test all_attributes() which includes own and inherited attributes
    assert class_child.all_attributes() == {attribute1, attribute2, attribute4}
    assert class_parent1.all_attributes() == {attribute1, attribute2}
    assert class_parent2.all_attributes() == {attribute1, attribute3}
    assert class_grandparent.all_attributes() == {attribute1}

def test_named_element_blank_spaces():
    # Test that names with spaces raise ValueError
    with pytest.raises(ValueError) as excinfo:
        named_element = NamedElement(name="element with spaces")
    assert "'element with spaces' is invalid. Name cannot contain spaces." in str(excinfo.value)

    # Test that names without spaces work fine
    named_element = NamedElement(name="element_without_spaces")
    assert named_element.name == "element_without_spaces"

def test_attribute_reassignment():
    attribute1: Property = Property(name="attr1", type=StringType)
    class1: Class = Class(name="Cls1", attributes={attribute1})
    class2: Class = Class(name="Cls2", attributes=set())

    # Reassign attribute1 to class2
    class2.attributes = {attribute1}

    assert attribute1 in class2.attributes
    assert attribute1 not in class1.attributes
    assert attribute1.owner == class2

    def test_package_initialization():
        # Create classes
        class1: Class = Class(name="Class1", attributes=set())
        class2: Class = Class(name="Class2", attributes=set())

        # Create associations
        aend1: Property = Property(name="end1", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
        aend2: Property = Property(name="end2", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
        association1: BinaryAssociation = BinaryAssociation(name="Association1", ends={aend1, aend2})

        aend3: Property = Property(name="end3", owner=None, type=class1, multiplicity=Multiplicity(0, 1))
        aend4: Property = Property(name="end4", owner=None, type=class2, multiplicity=Multiplicity(0, 1))
        association2: BinaryAssociation = BinaryAssociation(name="Association2", ends={aend3, aend4})

        # Create enumeration
        literal1: EnumerationLiteral = EnumerationLiteral(name="Literal1", owner=None)
        enumeration: Enumeration = Enumeration(name="Enumeration", literals={literal1})

        # Create package
        package: Package = Package(name="Package1", elements={class1, class2, association1, association2,enumeration})

        # Test get_classes method
        classes = package.get_classes()
        assert len(classes) == 2
        assert class1 in classes
        assert class2 in classes

        # Test get_associations method
        associations = package.get_associations()
        assert len(associations) == 2
        assert association1 in associations
        assert association2 in associations

        # Test get_enumerations method
        enumerations = package.get_enumerations()
        assert len(enumerations) == 1
        assert enumeration in enumerations

def test_domain_model_elements_recalculation():
    # Create types
    class1: Class = Class(name="Class1")
    class2: Class = Class(name="Class2")

    # Create associations
    aend1: Property = Property(name="end1", type=class1, multiplicity=Multiplicity(0, 1))
    aend2: Property = Property(name="end2", type=class2, multiplicity=Multiplicity(0, 1))
    association1: BinaryAssociation = BinaryAssociation(name="Association1", ends={aend1, aend2})

    # Create generalizations
    generalization1: Generalization = Generalization(general=class1, specific=class2)

    # Create packages
    package1: Package = Package(name="Package1", elements={class1, class2, association1, generalization1})

    # Create constraints
    constraint1: Constraint = Constraint(name="Constraint1", context=class1, expression="context Class1 inv: self.end1->notEmpty()", language="OCL")

    # Create domain model
    domain_model: DomainModel = DomainModel(
        name="TestDomainModel",
        types={class1, class2},
        associations={association1},
        generalizations={generalization1},
        packages={package1},
        constraints={constraint1}
    )

    # Verify elements property
    expected_elements = {class1, class2, association1, generalization1, package1, constraint1}
    assert domain_model.elements == (expected_elements | data_types)



import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.generators.pydantic_classes import PydanticGenerator

class RESTAPIGenerator(GeneratorInterface):
    """
    Rest_API_Generator is a class that implements the GeneratorInterface and is responsible for generating
    the Rest API domain model code based on the input B-UML model. This version of the generator allows
    specifying which HTTP methods (e.g., GET, POST, PATCH) should be included in the generated code. It can
    generate code for one or more specified methods, enabling more customizable API endpoint generation.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        http_methods (list): A list of strings representing the HTTP methods for which code should be generated.
                         Each element should be one of "GET", "POST", "PUT","PATCH","DELETE". This allows generating
                         only the parts of the API that are needed.
        backend (bool, optional): A boolean flag indicating whether the generator should generate code for a backend API.
        nested_creations (bool, optional): This parameter determines how entities are linked in the API request. 
                                            If set to True, both nested creations and linking by the ID of the entity 
                                            are enabled. If set to False, only the ID of the linked entity will be used.
                                            The default value is False.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """
    def __init__(self, model: DomainModel, http_methods: list = None, nested_creations: bool = False, backend: bool = False, port: int = None, output_dir: str = None):
        super().__init__(model, output_dir)
        allowed_methods = ["GET", "POST", "PUT", "PATCH", "DELETE"]
        if not http_methods:
            http_methods = allowed_methods
        else:
            http_methods = [method for method in http_methods if method in allowed_methods]
        self.http_methods = http_methods
        self.backend = backend
        self.nested_creations = nested_creations
        self.port = port

    def generate_requirements(self):
        """
        Generates requirements.txt file with necessary dependencies
        """
        requirements = [
            "fastapi>=0.68.0",
            "uvicorn>=0.15.0",
            "pydantic>=1.8.0",
            "typing-extensions>=4.0.0",
            "sqlalchemy>=1.4.0",
            "python-multipart>=0.0.5"
        ]
        
        file_path = self.build_generation_path(file_name="requirements.txt")
        with open(file_path, "w") as f:
            f.write("\n".join(requirements))

    def generate(self):
        """
        Generates Rest API model code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named rest_api.py and uses the Pydantic_Generator to generate
            the Pydantic classes
        """
        # Generate requirements.txt first
        self.generate_requirements()

        if self.backend:
            file_path = self.build_generation_path(file_name="main_api.py")
            templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
            env = Environment(loader=FileSystemLoader(templates_path),
                          trim_blocks=True, lstrip_blocks=True, extensions=['jinja2.ext.do'])
            template = env.get_template('backend_fast_api_template.py.j2')
            with open(file_path, mode="w") as f:
                generated_code = template.render(name=self.model.name, classes=self.model.classes_sorted_by_inheritance(),
                                             http_methods=self.http_methods, nested_creations=self.nested_creations, port=self.port)
                f.write(generated_code)
            print("Code generated in the location: " + file_path)

        else:
            pydantic_model = PydanticGenerator(model=self.model, backend=self.backend, nested_creations=self.nested_creations, output_dir=self.output_dir)
            pydantic_model.generate()
            
            file_path = self.build_generation_path(file_name="rest_api.py")
            templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
            env = Environment(loader=FileSystemLoader(templates_path),
                          trim_blocks=True, lstrip_blocks=True, extensions=['jinja2.ext.do'])
            template = env.get_template('fast_api_template.py.j2')
            with open(file_path, mode="w") as f:
                generated_code = template.render(classes=self.model.classes_sorted_by_inheritance(),
                                             http_methods=self.http_methods)
                f.write(generated_code)
            print("Code generated in the location: " + file_path)



import os
from besser.BUML.metamodel.gui import *
from besser.BUML.metamodel.structural import *
from jinja2 import Template, Environment, FileSystemLoader
from besser.generators import GeneratorInterface



##############################
#    SQL_Helper Generator
##############################

class FlutterSQLHelperGenerator(GeneratorInterface):

    """
    FlutterSQLHelperGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the Flutter SQL helper code based on the input B-UML model.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    TYPES = {
        "int": "int?",
        "str": "String",
        "float": "float8",
        "bool": "bool",
        "time": "time",
        "date": "date",
        "datetime": "timestamp",
        "timedelta": "interval",
    }

    def __init__(self, model: DomainModel, output_dir: str = None):
       super().__init__(model, output_dir)
       

    def generate(self):

        """
        Generates Flutter SQL helper code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but stores the generated code as a file named sql_helper.dart.
        """

        copy_model: DomainModel = self.model

        for cls in copy_model.get_classes():
                 attr_list = list(cls.attributes)
                 cls.attributes = attr_list
        
        for cls in copy_model.get_classes():
            for atr in cls.attributes:
              print(cls.name + " ::  "+ atr.name)

        
        file_path = self.build_generation_path(file_name="sql_helper.dart")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('flutterCodeGeneratorSqlHelperFile.py.j2')
        with open(file_path, mode="w") as f:
            generated_code = template.render(BUMLClasses= copy_model.get_classes(), model=copy_model, types=self.TYPES)
            f.write(generated_code)
            print("Code generated in the location: " + file_path)
        



##############################
#    Main Dart Generator
##############################

class FlutterMainDartGenerator(GeneratorInterface):

    """
    FlutterMainDartGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the main Dart code for a Flutter application based on the input B-UML and GUI models.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        application (Application): An instance of the Application class representing the GUI model.
        main_page (Screen): An instance of the Screen class representing the main page of the Flutter application.
        module (Module): An instance of the Module class representing the module of the Flutter application.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    def __init__(self, model: DomainModel, application: Application,  main_page: Screen, module: Module = None, output_dir: str = None):
        super().__init__(model, output_dir)
        self.application: Application = application
        self.main_page: Screen = main_page
        self.module: Module = module

    @property
    def application(self) -> Application:
        """Application: Get the instance of the Application class representing the GUI model."""
        return self.__application

    @application.setter
    def application(self, application: Application):
        """Application: Set the instance of the Application class representing the GUI model."""
        self.__application = application

    @property
    def main_page(self) -> Screen:
        """Screen: Get the instance of the Screen class representing the main page of the Flutter application."""
        return self.__main_page

    @main_page.setter
    def main_page(self, main_page: Screen):
        """Screen: Set the instance of the Screen class representing the main page of the Flutter application."""
        self.__main_page = main_page

    @property
    def module(self) -> Module:
        """Module: Get the instance of the Module class representing the module of the Flutter application."""
        return self.__module

    @module.setter
    def module(self, module: Module):
        """Module: Set the instance of the Module class representing the module of the Flutter application."""
        self.__module = module
    
    @staticmethod
    def is_Button(value):
        """Check if the given value is an instance of Button class."""
        return isinstance(value, Button)

    @staticmethod
    def is_List(value):
        """Check if the given value is an instance of DataList class."""
        return isinstance(value, DataList)

    @staticmethod
    def is_ModelElement(value):
        """Check if the given value is an instance of ModelElement class."""
        return isinstance(value, ModelElement)

    def generate(self):

        """
        Generates the main Dart code for a Flutter application based on the provided B-UML and GUI models and saves it
        to the specified output directory. 
        If the output directory was not specified, the code generated will be
        stored in the <current directory>/output folder.

        Returns:
            None, but stores the generated code as a file named main.dart.
        """

        file_path = self.build_generation_path(file_name="main.dart")
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('flutterCodeGeneratorMainFile.py.j2')
        env.tests['is_Button'] = self.is_Button
        env.tests['is_List'] = self.is_List
        env.tests['is_ModelElement'] = self.is_ModelElement
        if self.module is None:
          # User did not specify a module, so select the first module from the set of modules
          self.module = next(iter(self.application.modules))

        screens = self.module.screens
        screens.remove(self.main_page)
        for scr in screens:
              print(scr.name + " ::  ")
  
        with open(file_path, mode="w") as f:
            generated_code = template.render(
                app=self.application,
                screens=screens,
                screen=self.main_page,
                BUMLClasses=self.model.get_classes(),
                model=self.model,
                associations=self.model.associations
            )
            f.write(generated_code)
            print("Code generated in the location: " + file_path)
       

##############################
#    pubspec Generator
##############################


class FlutterPubspecGenerator(GeneratorInterface):

    """
    FlutterPubspecGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the pubspec.yaml file for a Flutter application based on the input GUI model.

    Args:
        application (Application): An instance of the Application class representing the GUI model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    def __init__(self, application: Application, output_dir: str = None):
        super().__init__(output_dir)
        self.application: Application = application


    @property
    def application(self) -> Application:
        """Application: Get the instance of the Application class representing the GUI model."""
        return self.__application

    @application.setter
    def application(self, application: Application):
        """Application: Set the instance of the Application class representing the GUI model."""
        self.__application = application


    def generate(self):

        """
        Generates the pubspec.yaml file for a Flutter application based on the provided GUI model and saves it
        to the specified output directory. If the output directory was not specified, the code generated will be
        stored in the <current directory>/output folder.

        Returns:
            None, but stores the generated code as a file named pubspec.yaml.
        """
        
        file_path = self.build_generation_path(file_name="pubspec.yaml")
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('flutterCodeGeneratorPubspecFile.py.j2')
        with open(file_path, mode="w") as f:
            generated_code = template.render(
                app=self.application
            )
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



##############################
#    Flutter Generator
##############################

class FlutterGenerator(GeneratorInterface):
        
    """
        A class that represents a Flutter code generator.

        This class extends the GeneratorInterface, which is an interface or base class for code generators.

        Args:
            model: An object representing the B-UML model.
            application: An object representing the GUI model.
            main_page: An object representing the main page of the Flutter application.
            module: An object representing the module of the Flutter application.
            output_dir: The output directory where the generated code will be saved. (optional)
     """
    def __init__(self, model, application, main_page, module=None, output_dir=None):
        super().__init__(model, output_dir)

        self.application = application
        self.main_page = main_page
        self.module = module


    def generate(self):

        """
        Generates the Flutter code based on the provided models.

        This method creates instances of the FlutterSQLHelperGenerator, FlutterMainDartGenerator,
        and FlutterPubspecGenerator classes. It then calls the generate() method on each of them
        to generate the SQL helper code, main Dart code, and pubspec.yaml file, respectively.

        Returns:
            None
            None, but store the generated code as files main.dart, sql_helper.dart, and pubspec.yaml.
        """


        sql_helper_generator = FlutterSQLHelperGenerator(model=self.model, output_dir=self.output_dir)
        sql_helper_generator.generate()

        main_dart_generator = FlutterMainDartGenerator(model=self.model, application=self.application, main_page=self.main_page, module=self.module, output_dir=self.output_dir)
        main_dart_generator.generate()

        pubspec_generator = FlutterPubspecGenerator(application=self.application, output_dir=self.output_dir)
        pubspec_generator.generate()


import pytest

from besser.BUML.metamodel.feature_model.feature_model import FeatureModel, Feature, FeatureValue, FeatureGroup, \
    MANDATORY, OPTIONAL, ALTERNATIVE, OR

INF = 999


def test_feature_value():
    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='int', values=[1, 2, 3], min=1)
    assert "Invalid arguments" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='int', values=[1, 2, 3], max=1)
    assert "Invalid arguments" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='int', values=[1, 2, 3], min=1, max=1)
    assert "Invalid arguments" in str(excinfo.value)

    # This should work
    feature_value = FeatureValue(t='int', values=[1, 2, 3])
    feature_value = FeatureValue(t='float', values=[1.0, 2.0, 3.0])
    feature_value = FeatureValue(t='str', values=['a', 'b', 'c'])

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='str', values=[1, 2, 3])
    assert "Value must be a string" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='int', values=['a', 'b', 'c'])
    assert "Value must be an integer" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_value = FeatureValue(t='float', values=['a', 'b', 'c'])
    assert "Value must be a float" in str(excinfo.value)


def test_feature_cardinality():
    # This should work
    feature = Feature('feature', min=1, max=1)
    feature = Feature('feature', min=1, max=2)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature = Feature('feature', min=2, max=1)
    assert f'Error in {feature.name}: 0 < min < max' in str(excinfo.value)


def test_mandatory():
    feature1 = Feature('feature1')
    feature2 = Feature('feature2')
    child = Feature('child')

    # This should work
    feature1.mandatory(child)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature2.mandatory(child)
    assert f'Feature {child.name} cannot be a child of {feature2.name}. It has feature {child.parent.name} as parent.' in str(excinfo.value)


def test_optional():
    feature1 = Feature('feature1')
    feature2 = Feature('feature2')
    child = Feature('child')

    # This should work
    feature1.optional(child)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature2.optional(child)
    assert f'Feature {child.name} cannot be a child of {feature2.name}. It has feature {child.parent.name} as parent.' in str(
        excinfo.value)


def test_alternative():
    feature1 = Feature('feature1')
    feature2 = Feature('feature2')
    child1 = Feature('child1')
    child2 = Feature('child2')

    # This should work
    feature1.alternative([child1, child2])

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature2.alternative([child1, child2])
    assert f'Feature {child1.name} cannot be a child of {feature2.name}. It has feature {child1.parent.name} as parent.' in str(excinfo.value)


def test_or():
    feature1 = Feature('feature1')
    feature2 = Feature('feature2')
    child1 = Feature('child1')
    child2 = Feature('child2')

    # This should work
    feature1.or_([child1, child2])

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature2.or_([child1, child2])
    assert f'Feature {child1.name} cannot be a child of {feature2.name}. It has feature {child1.parent.name} as parent.' in str(excinfo.value)


def test_feature_group():
    feature1 = Feature('feature1')
    feature2 = Feature('feature2')

    # This should work
    feature_group = FeatureGroup(kind=MANDATORY, features=[feature1])
    feature_group = FeatureGroup(kind=OPTIONAL, features=[feature1])
    feature_group = FeatureGroup(kind=ALTERNATIVE, features=[feature1, feature2])
    feature_group = FeatureGroup(kind=OR, features=[feature1, feature2])

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_group = FeatureGroup(kind=MANDATORY, features=[feature1, feature2])
    assert f'{MANDATORY} has more than 1 feature' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_group = FeatureGroup(kind=OPTIONAL, features=[feature1, feature2])
    assert f'{OPTIONAL} has more than 1 feature' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_group = FeatureGroup(kind=ALTERNATIVE, features=[feature1])
    assert f'{ALTERNATIVE} has less than 2 features' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # This should not work
        feature_group = FeatureGroup(kind=OR, features=[feature1])
    assert f'{OR} has less than 2 features' in str(excinfo.value)


def test_duplicate_feature_model():
    feature_model1 = (FeatureModel('feature_model').root(
        Feature('feature1')
        .mandatory(Feature('feature2'))
        .optional(Feature('feature2'))
        .alternative([
            Feature('feature3'),
            Feature('feature4')
        ])
        .or_([
            Feature('feature5'),
            Feature('feature6')
            .mandatory(Feature('feature7', min=2, max=10, value=FeatureValue('int', values=[1, 2, 3])))
        ])
    ))
    feature_model2 = FeatureModel('feature_model').root(feature_model1.duplicate())
    assert feature_model1 == feature_model2
    feature_model2.root_feature.mandatory(Feature('feature8'))
    assert feature_model1 != feature_model2
